#include "match.h"
#include "machine_state.h"
#include "match_struct.h"
#include "regex_struct.h"
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

typedef struct RegexMatcher_s {
	RegexMatchThreadGroup top_group;
	DynamicArray visitor_stack;
	size_t depth;

	RegexMatchThread finished_thread;
} RegexMatcher;

static int RegexMatchThread_New(RegexMatchThread* thread, StringView string, size_t index, Regex* regex)
{
	thread->string = string;
	thread->left   = StringView_Slice(string, 0, index);
	thread->right  = StringView_Slice(string, index, string.length);
	thread->regex  = regex;

	thread->machine_head = regex->head;

	thread->match.match = StringView_Slice(string, index, index);
	int da_code = DynamicArray_Create(
		&thread->match.captures,
		sizeof(RegexCapture),
		8,
		regex->machine_memory.allocator
	);
	if (da_code != EXIT_SUCCESS) return da_code;

	da_code = DynamicArray_Create(
		&thread->repeats,
		sizeof(RegexRepeatStackEntry),
		8,
		regex->machine_memory.allocator
	);
	if (da_code != EXIT_SUCCESS) return da_code;

	return EXIT_SUCCESS;
}

static int _RegexMatch_Clone(RegexCapture* new, RegexCapture* orig, allocator_t* allocator)
{
	char* new_name = Allocator_Allocate(allocator, strlen(orig->name) + 1);
	if (new_name == NULL) {
		return ENOMEM;
	}

	new->name = new_name;

	strcpy(new->name, orig->name);

	return EXIT_SUCCESS;
}

static int RegexMatchThread_Clone(RegexMatchThread* new, RegexMatchThread* old)
{
	*new = *old;

	int clone_code;
	if (new->regex->options & REGEX_MATCH_STRDUP) {
		// strdup the capture names
		clone_code = DynamicArray_DeepClone(
			&new->match.captures,
			&old->match.captures,
			(DynamicArrayCloneFunction) _RegexMatch_Clone,
			old->match.captures.allocator
		);
	} else {
		clone_code = DynamicArray_Clone(&new->match.captures, &old->match.captures);
	}
	if (clone_code == ECANCELED)
		clone_code = ENOMEM;
	if (clone_code != EXIT_SUCCESS) {
		return clone_code;
	}

	// Array was initialized
	if (old->repeats.capacity != 0) {
		DynamicArray_Clone(&new->repeats, &old->repeats);
	}

	return EXIT_SUCCESS;
}

static void RegexMatchThread_Del(RegexMatchThread* thread)
{
	if (thread->regex->options & REGEX_MATCH_STRDUP) {
		DYNAMICARRAY_FOREACH(thread->match.captures, i) {
			RegexCapture* capture = DynamicArray_GetPointer(&thread->match.captures, i);
			Allocator_Free(
				thread->match.captures.allocator,
				capture->name,
				strlen(capture->name) + 1
			);
		}
	}
	DynamicArray_Destroy(&thread->match.captures);
	DynamicArray_Destroy(&thread->repeats);
	memset(thread, 0, sizeof(*thread));
}

static int _HandleAccept(RegexMatchThread* thread, bool* discard)
{
	RegexMachineStateAccept* accept = (RegexMachineStateAccept*) thread->machine_head;

	// Matching
	if (StringView_StartsWith(thread->right, accept->sequence)) {
		thread->right = StringView_Slice(
			thread->right,
			accept->sequence.length,
			thread->right.length
		);
		thread->left.length += accept->sequence.length;
		thread->match.match.length += accept->sequence.length;
	} else {
		*discard = true;
	}

	return EXIT_SUCCESS;
}

static int _HandleGroup(RegexMatchThread* thread)
{
	RegexMachineStateGroup* group = (RegexMachineStateGroup*) thread->machine_head;

	if (RegexMatch_HaveNumberedCapture(&thread->match, group->number)) {
		// End of group, store match
		RegexCapture* capture = RegexMatch_GetNumberedCapture(&thread->match, group->number);

		if (StringView_Equal(capture->match, STRINGVIEW_NONE)) {
			// Match is finished after being reset or started
			capture->match_end = thread->left.length;
			capture->match = StringView_Slice(thread->left, capture->match_start, capture->match_end);

		} else {
			// Recapture start: e.g. "(a)+" on string "aa", reset match
			capture->match_start = thread->left.length;
			capture->match = STRINGVIEW_NONE;
		}

	} else {
		// Start of group, start match
		RegexCapture* capture = RegexMatch_MakeCapture(thread, group->name, group->number);
		if (capture == NULL) {
			return ENOMEM;
		}

		capture->match_start = thread->left.length;
		capture->match = STRINGVIEW_NONE;
	}

	return EXIT_SUCCESS;
}

static int _HandleRepeat(RegexMatchThread* thread)
{
	RegexMachineStateRepeat* repeat = (RegexMachineStateRepeat*) thread->machine_head;
	RegexRepeatStackEntry* top_entry = DynamicArray_GetPointer(&thread->repeats, thread->repeats.reserved - 1);

	if (top_entry->node == repeat) {
		// At least first iteration, maybe even repetitions
	} else {
		// Before first iteration, not matched once

		DynamicArray_AppendEmpty(&thread->repeats, (void**) &top_entry);
		if (top_entry == NULL) {
			return ENOMEM;
		}

		top_entry->node = repeat;
		top_entry->current = 0;

		// possibly create more threads
	}

	return EXIT_SUCCESS;
}

union ThreadOrGroup {
	RegexMatchThread* thread;
	RegexMatchThreadGroup* group;
};

struct GroupVisitor {
	// Cast these into the types
	enum RegexMatchThreadType type;
	unsigned int position;
	union ThreadOrGroup as;
};

static int RegexMatchThreadGroup_Create(RegexMatchThreadGroup* group, size_t number, allocator_t* allocator)
{
	memset(group, 0, sizeof(*group));
	group->number = number;
	group->threads.allocator = allocator;
	return EXIT_SUCCESS;
}

static int _FindThreadLinear(RegexMatchThread* other, RegexMatchThread* find)
{
	if (other->number < find->number) {
		return 1;
	} else if (other->number > find->number) {
		return -1;
	} else {
		return 0;
	}
}

static RegexMatchThread* _RegexMatchThreadGroup_NewThread(RegexMatchThreadGroup* group)
{
	if (DynamicArray_GetLength(&group->threads) == 0) {
		int create_code = DynamicArray_Create(
			&group->threads,
			sizeof(RegexMatchThread),
			4,
			group->threads.allocator
		);
		if (create_code) return NULL;
	}

	RegexMatchThread* pointer;
	DynamicArray_AppendEmpty(&group->threads, (void**) &pointer);

	return pointer;
}

static RegexMatchThreadGroup* RegexMatchThreadGroup_NewGroup(RegexMatchThreadGroup* parent_group)
{
	if (DynamicArray_GetLength(&parent_group->subgroups) == 0) {
		int create_code = DynamicArray_Create(
			&parent_group->subgroups,
			sizeof(RegexMatchThreadGroup),
			4,
			parent_group->threads.allocator
		);
		if (create_code) return NULL;
	}

	RegexMatchThreadGroup* pointer;
	DynamicArray_AppendEmpty(&parent_group->subgroups, (void**) &pointer);

	return pointer;
}


static void _RegexMatchThreadGroup_ForgetThread(RegexMatchThreadGroup* group, RegexMatchThread* thread)
{
	size_t thread_index = DynamicArray_FindFunctionLinear(
		&group->threads,
		(DynamicArrayLinearFindFunction) _FindThreadLinear,
		thread
	);

	DynamicArray_Remove(&group->threads, thread_index);

	return;
}

static void RegexMatchThreadGroup_Destroy2(RegexMatchThreadGroup* group)
{
	DynamicArray_Destroy(&group->threads);
	DynamicArray_Destroy(&group->subgroups);
	return;
}

static void RegexMatchThreadGroup_Destroy(DynamicArray* visitors)
{

	struct GroupVisitor* current_visitor;
	current_visitor = DynamicArray_GetPointer(
		visitors,
		visitors->reserved - 1
	);
	RegexMatchThreadGroup* group = current_visitor->as.group;

	if (DynamicArray_GetLength(&group->threads) != 0) {
		// Append first sub thread to visitor stack

		struct GroupVisitor* new_visitor;
		DynamicArray_AppendEmpty(visitors, (void**) &new_visitor);
		RegexMatchThread* thread = DynamicArray_GetPointer(
			&group->threads,
			0
		);
		new_visitor->type = REGEXMATCHTHREADTYPE_THREAD;
		new_visitor->as.thread = thread;

		DynamicArray_RemoveFast(&group->threads, 0);
	} else if (DynamicArray_GetLength(&group->subgroups) != 0) {
		// Append first sub group to visitor stack

		struct GroupVisitor* new_visitor;
		DynamicArray_AppendEmpty(visitors, (void**) &new_visitor);
		RegexMatchThreadGroup* subgroup = DynamicArray_GetPointer(
			&group->subgroups,
			0
		);
		new_visitor->type = REGEXMATCHTHREADTYPE_GROUP;
		new_visitor->as.group = subgroup;

		DynamicArray_RemoveFast(&group->subgroups, 0);

	} else {
		// Actually destroy the group, also pop the visitor
		RegexMatchThreadGroup_Destroy2(group);
		DynamicArray_RemoveFast(visitors, visitors->reserved - 1);
	}
	return;
}

static int _RegexMatcher_Create(RegexMatcher* matcher, Regex* regex, StringView string, size_t start)
{
	int group_code = RegexMatchThreadGroup_Create(
		&matcher->top_group,
		0,
		regex->machine_memory.allocator
	);
	if (group_code) {
		return group_code;
	}

	int stack_code = DynamicArray_Create(
		&matcher->visitor_stack,
		sizeof(struct GroupVisitor),
		16,
		regex->machine_memory.allocator
	);
	if (stack_code) {
		RegexMatchThreadGroup_Destroy2(&matcher->top_group);
		return stack_code;
	}

	RegexMatchThread* root_thread = _RegexMatchThreadGroup_NewThread(&matcher->top_group);
	if (root_thread == NULL) {
		RegexMatchThreadGroup_Destroy2(&matcher->top_group);
		DynamicArray_Destroy(&matcher->visitor_stack);
		return ENOMEM;
	}
	if (RegexMatchThread_New(root_thread, string, start, regex)) {
		RegexMatchThreadGroup_Destroy2(&matcher->top_group);
		DynamicArray_Destroy(&matcher->visitor_stack);
		return ENOMEM;
	}

	matcher->depth = 1;

	memset(&matcher->finished_thread, 0, sizeof(matcher->finished_thread));

	return EXIT_SUCCESS;
}

static void _RegexMatcher_Destroy(RegexMatcher* matcher)
{
	RegexMatchThreadGroup_Destroy2(&matcher->top_group);
	DynamicArray_Destroy(&matcher->visitor_stack);
	memset(matcher, 0, sizeof(*matcher));
}

static int _NewRegexChild(RegexMatchThread** parent, DynamicArray* visitors, size_t* depth, RegexMachineStateBase* new_head)
{
	RegexMatchThread* child;
	struct GroupVisitor* parent_visitor = DynamicArray_GetPointer(
		visitors,
		visitors->reserved - 2
	);
	if (parent_visitor->type != REGEXMATCHTHREADTYPE_GROUP) {
		return EXIT_FAILURE;
	}

	RegexMatchThreadGroup* parent_group = parent_visitor->as.group;
	if (parent_group->number == (*parent)->number) {
		// group is owned by parent
		DynamicArray_AppendEmpty(
			&parent_group->threads,
			(void**)&child
		);
	} else {
		// New subgroup, move parent thread

		// Enlarge visitor array to provide enough space for visiting
		if (visitors->reserved - 1 == *depth) {
			*depth += 1;
			if (visitors->capacity == *depth) {
				int resize = DynamicArray_Resize(
					visitors,
					visitors->capacity * 2
				);
				if (resize) return resize;
			}
		}

		RegexMatchThreadGroup* subgroup;
		subgroup = RegexMatchThreadGroup_NewGroup(parent_group);

		if (subgroup == NULL) {
			return ENOMEM;
		}

		// make a new group
		RegexMatchThreadGroup_Create(
			subgroup,
			(*parent)->number,
			(*parent)->regex->machine_memory.allocator
		);

		// acquire new thread memory
		RegexMatchThread* nparent = _RegexMatchThreadGroup_NewThread(
			subgroup
		);
		child = _RegexMatchThreadGroup_NewThread(subgroup);

		// memowy
		if (nparent == NULL || child == NULL) {
			RegexMatchThreadGroup_Destroy2(subgroup);
			DynamicArray_RemoveFast(
				&parent_group->subgroups,
				parent_group->subgroups.reserved - 1
			);
			return ENOMEM;
		}

		// copy contents
		*nparent = **parent;

		// update reference
		*parent = nparent;
		parent_group = subgroup;

		// remove from old group
		_RegexMatchThreadGroup_ForgetThread(parent_group, *parent);
		parent_visitor->position--;
	}

	if (child == NULL) {
		return ENOMEM;
	}

	int clone_code = RegexMatchThread_Clone(child, *parent);
	if (clone_code) {
		return clone_code;
	}

	child->machine_head = new_head;
	child->number = (*parent)->number
			+ DynamicArray_GetLength(&parent_group->threads) - 1;

	return EXIT_SUCCESS;
}

static int _HandleOption(RegexMatcher* matcher)
{
	struct GroupVisitor* visitor = DynamicArray_GetPointer(
		&matcher->visitor_stack,
		DynamicArray_GetLength(&matcher->visitor_stack)
	);
	RegexMatchThread* thread = visitor->as.thread;

	RegexMachineStateBase* first_head = thread->machine_head;
	{
		RegexMachineStateOption* opt = (void*) first_head;
		thread->machine_head = &opt->next_option->base;
	}

	while (thread->machine_head->type == REGEXMACHINESTATETYPE_OPTION) {
		RegexMachineStateBase* option_head = thread->machine_head->next;

		int child_code = _NewRegexChild(
			&thread,
			&matcher->visitor_stack,
			&matcher->depth,
			option_head->next
		);
		if (child_code) {
			return child_code;
		}
	}

	// will be auto-advanced outside of this function
	thread->machine_head = first_head;

	return EXIT_SUCCESS;
}

static void _TryFinishThread(RegexMatcher* matcher, RegexMatchThread* thread)
{
	struct GroupVisitor* parent_visitor = DynamicArray_GetPointer(
		&matcher->visitor_stack,
		DynamicArray_GetLength(&matcher->visitor_stack) - 2
	);
	RegexMatchThreadGroup* parent_group = parent_visitor->as.group;
	size_t thread_index = DynamicArray_FindFunctionLinear(
		&parent_group->threads,
		(DynamicArrayLinearFindFunction) _FindThreadLinear,
		thread
	);

	if (thread_index == 0
		&& DynamicArray_GetLength(&matcher->visitor_stack) == 2) {
		// Only the first thread in the top-most group is finishable
		// All other threads have a lower priority, this is because
		// of the order they have to be matched in.
		matcher->finished_thread = *thread;
		_RegexMatchThreadGroup_ForgetThread(parent_group, thread);
		matcher->visitor_stack.reserved = 0;
	}
}

static int _AdvanceThread(RegexMatcher* matcher)
{
	struct GroupVisitor* visitor = DynamicArray_GetPointer(
		&matcher->visitor_stack,
		DynamicArray_GetLength(&matcher->visitor_stack) - 1
	);
	RegexMatchThread* thread = visitor->as.thread;
	int code;
	bool discard;

	switch (thread->machine_head->type) {
	case REGEXMACHINESTATETYPE_ACCEPT:
		code = _HandleAccept(thread, &discard);
		break;
	case REGEXMACHINESTATETYPE_GROUP:
		code = _HandleGroup(thread);
		break;
	case REGEXMACHINESTATETYPE_OPTION:
		code = _HandleOption(matcher);
		break;
	case REGEXMACHINESTATETYPE_REPEAT:
		code = _HandleRepeat(thread);
		break;
	default:
		return EXIT_FAILURE;
	}

	if (code == EXIT_SUCCESS) {
		thread->machine_head = thread->machine_head->next;
	}

	if (thread->machine_head == NULL) {
		_TryFinishThread(matcher, thread);
	}

	// Remove from visitor stack
	DynamicArray_Remove(&matcher->visitor_stack, matcher->visitor_stack.reserved - 1);

	return EXIT_SUCCESS;
}

static void _HandleDestroyVisitor(DynamicArray* visitors, struct GroupVisitor* current_visitor)
{
	switch (current_visitor->type) {
		case REGEXMATCHTHREADTYPE_GROUP:
			RegexMatchThreadGroup_Destroy(visitors);
			break;
		case REGEXMATCHTHREADTYPE_THREAD:
			RegexMatchThread_Del(current_visitor->as.thread);
			break;
	}
}

static void _DestroyThreadHierarchy(DynamicArray* visitors, RegexMatchThreadGroup* group)
{
	visitors->reserved = 0;
	{
		struct GroupVisitor* zero_visitor;
		DynamicArray_AppendEmpty(visitors, (void**) &zero_visitor);
		zero_visitor->type = REGEXMATCHTHREADTYPE_GROUP;
		zero_visitor->as.group = group;
	}

	while (DynamicArray_GetLength(visitors) != 0) {
		struct GroupVisitor* current_visitor;
		current_visitor = DynamicArray_GetPointer(
			visitors,
			visitors->reserved - 1
		);

		_HandleDestroyVisitor(visitors, current_visitor);
	}

	return;
}

static int _AdvanceGroup(RegexMatcher* matcher)
{
	struct GroupVisitor* visitor = DynamicArray_GetPointer(
		&matcher->visitor_stack,
		DynamicArray_GetLength(&matcher->visitor_stack) - 1
	);
	RegexMatchThreadGroup* group = visitor->as.group;
	visitor->position += 1;

	if (DynamicArray_GetLength(&group->threads) > visitor->position) {
		struct GroupVisitor* child_visitor;
		DynamicArray_AppendEmpty(
			&matcher->visitor_stack,
			(void**) &child_visitor
		);
		child_visitor->type = REGEXMATCHTHREADTYPE_THREAD;
		child_visitor->as.thread = DynamicArray_GetPointer(&group->threads, visitor->position);
	} else {
		// calculate position in subgroup array
		unsigned int position = visitor->position
			- DynamicArray_GetLength(&group->threads);

		if (DynamicArray_GetLength(&group->subgroups) == position) {
			// Remove this group
			DynamicArray_Remove(
				&matcher->visitor_stack,
				matcher->visitor_stack.reserved - 1
			);
		} else {
			// Add subgroup
			struct GroupVisitor* sgroup_visitor;
			DynamicArray_AppendEmpty(
				&matcher->visitor_stack,
				(void**) &sgroup_visitor
			);
			sgroup_visitor->type = REGEXMATCHTHREADTYPE_GROUP;
			sgroup_visitor->position = -1;
			sgroup_visitor->as.group = DynamicArray_GetPointer(
				&group->subgroups,
				position
			);
		}
	}

	return EXIT_SUCCESS;
}

static int _HandleAdvanceVisitor(RegexMatcher* matcher)
{
	struct GroupVisitor* current_visitor = DynamicArray_GetPointer(
		&matcher->visitor_stack,
		DynamicArray_GetLength(&matcher->visitor_stack) - 1
	);

	int advance_code = EXIT_SUCCESS;
	switch (current_visitor->type) {
		case REGEXMATCHTHREADTYPE_THREAD:
			advance_code = _AdvanceThread(matcher);
			break;
		case REGEXMATCHTHREADTYPE_GROUP:
			advance_code = _AdvanceGroup(matcher);
			break;
	}

	return advance_code;
}

static int _WalkThreads(RegexMatcher* matcher)
{
	{

	struct GroupVisitor* zero_visitor;
	DynamicArray_AppendEmpty(
		&matcher->visitor_stack,
		(void**) &zero_visitor
	);
	zero_visitor->type = REGEXMATCHTHREADTYPE_GROUP;
	zero_visitor->as.group = &matcher->top_group;
	zero_visitor->position = -1;

	}

	while (DynamicArray_GetLength(&matcher->visitor_stack) != 0) {
		int advance_code = _HandleAdvanceVisitor(matcher);
		if (advance_code) return advance_code;
	}
	return EXIT_SUCCESS;
}

static bool _HaveFinishedThread(RegexMatcher* matcher)
{
	return matcher->finished_thread.regex != NULL;
}

static bool _HaveThreadsRunning(RegexMatcher* matcher)
{
	return matcher->top_group.subgroups.reserved != 0
		|| matcher->top_group.threads.reserved != 0;
}

int Regex_MatchHere(Regex* regex, StringView string, size_t start, RegexMatch* match)
{
	int return_code = EXIT_SUCCESS;

	RegexMatcher matcher;
	if (_RegexMatcher_Create(&matcher, regex, string, start))	{
		return EXIT_FAILURE;
	}

	while (! _HaveFinishedThread(&matcher)
		&& _HaveThreadsRunning(&matcher)) {
		_WalkThreads(&matcher);
	}

	if (_HaveFinishedThread(&matcher)) {
		*match = matcher.finished_thread.match;
		memset(
			&matcher.finished_thread.match,
			0,
			sizeof(matcher.finished_thread.match)
		);
		RegexMatchThread_Del(&matcher.finished_thread);
	}

	_DestroyThreadHierarchy(&matcher.visitor_stack, &matcher.top_group);
	_RegexMatcher_Destroy(&matcher);

	return return_code;
}

RegexCapture* RegexMatch_MakeCapture(RegexMatchThread* thread, char* name, size_t number)
{
	RegexMatch* match = &thread->match;
	RegexCapture* capture;
	DynamicArray_AppendEmpty(&match->captures, (void**) &capture);
	if (capture == NULL) {
		return NULL;
	}

	memset(capture, 0, sizeof(*capture));
	capture->number = number;
	if (name == NULL) {
		capture->name = NULL;
	} else {
		if (thread->regex->options & REGEX_MATCH_STRDUP) {
			char* name_copy = Allocator_Allocate(match->captures.allocator, strlen(name) + 1);
			if (name_copy == NULL) {
				DynamicArray_RemoveFast(&match->captures, match->captures.reserved - 1);
				return NULL;
			}
			strcpy(name_copy, name);
			capture->name = name_copy;
		} else {
			// Unowned reference
			capture->name = name;
		}
	}

	return capture;
}
